from __future__ import annotations

from typing import Any, cast

import sqlalchemy as sa
from sqlalchemy.engine import Engine

from ckan import model

from ckanext.datastore.backend.postgres import get_read_engine

from ckanext.collection import shared, types


class DbConnection(types.BaseDbConnection, shared.Domain[types.TDbCollection]):
    engine: Engine = shared.configurable_attribute()

    @property
    def inspector(self):
        return sa.inspect(self.engine)


class UrlDbConnection(DbConnection[types.TDbCollection]):
    url: str = shared.configurable_attribute()
    engine_options: dict[str, Any] = shared.configurable_attribute(
        default_factory=lambda self: {},
    )

    def __init__(self, obj: types.TDbCollection, /, **kwargs: Any):
        super().__init__(obj, engine=None, **kwargs)
        self.engine = sa.create_engine(self.url, **self.engine_options)


class CkanDbConnection(DbConnection[types.TDbCollection]):
    engine: Engine = shared.configurable_attribute(
        default_factory=lambda self: cast(Engine, model.meta.engine),
    )


class DatastoreDbConnection(DbConnection[types.TDbCollection]):
    engine: Engine = shared.configurable_attribute(
        default_factory=lambda self: get_read_engine(),
    )



from __future__ import annotations

from werkzeug.utils import import_string

import ckan.plugins.toolkit as tk

from ckanext.collection import types

CONFIG_ANNONYMOUS = "ckanext.collection.auth.anonymous_collections"
CONFIG_AUTHENTICATED = "ckanext.collection.auth.authenticated_collections"
CONFIG_INCLUDE_ASSET = "ckanext.collection.include_htmx_asset"
CONFIG_INIT_MODULES = "ckanext.collection.htmx_init_modules"


def anonymous_collections() -> list[str]:
    return tk.config[CONFIG_ANNONYMOUS]


def authenticated_collections() -> list[str]:
    return tk.config[CONFIG_AUTHENTICATED]


def include_htmx_asset() -> bool:
    return tk.config[CONFIG_INCLUDE_ASSET]


def htmx_init_modules() -> bool:
    return tk.config[CONFIG_INIT_MODULES]


def serializer(format: str) -> type[types.BaseSerializer] | None:
    value = tk.config.get(f"ckanext.collection.export.{format}.serializer")
    if value:
        return import_string(value, silent=True)
    return None



from __future__ import annotations

import csv
import json
from typing import Any

import pytest

from ckanext.collection.utils import Collection, serialize
from ckanext.collection.utils.data import StaticData


class StaticCollection(Collection):
    DataFactory = StaticData


@pytest.fixture()
def collection():
    return StaticCollection(
        "",
        {},
        data_settings={
            "data": [
                {"name": "a", "age": 1},
                {"name": "b", "age": 2},
            ],
        },
        columns_settings={
            "names": ["name", "age"],
        },
    )


class TestSerializer:
    def test_value_serializers(self, faker):
        record = {"name": faker.word(), "age": faker.pyint(), "gender": None}
        collection = StaticCollection(
            "",
            {},
            columns_settings={
                "serializers": {"name": [("up", {})], "age": [("xN", {"scale": 10})]},
            },
            serializer_settings={
                "value_serializers": {
                    "up": lambda value, *_: value.upper(),
                    "xN": lambda value, options, *_: value * options["scale"],
                },
            },
        )
        assert collection.serializer.serialize_value(
            record["name"],
            "name",
            record,
        ) == str.upper(record["name"])
        assert (
            collection.serializer.serialize_value(record["age"], "age", record)
            == record["age"] * 10
        )
        assert (
            collection.serializer.serialize_value(record["gender"], "gender", record)
            == record["gender"]
        )


class TestCsvSerializer:
    def test_output(self, collection: StaticCollection):
        serializer = serialize.CsvSerializer(collection)

        output = serializer.serialize().strip()
        nl = csv.get_dialect("excel").lineterminator
        expected_output = nl.join(["name,age", "a,1", "b,2"])

        assert output == expected_output


class TestJsonlSerializer:
    def test_output(self, collection: StaticCollection):
        serializer = serialize.JsonlSerializer(collection)

        output = serializer.serialize().strip()
        nl = "\n"
        expected_output = nl.join(map(json.dumps, collection.data))

        assert output == expected_output


class TestJsonSerializer:
    def test_output(self, collection: StaticCollection):
        serializer = serialize.JsonSerializer(collection)

        output = serializer.serialize().strip()
        expected_output = json.dumps(list(collection.data))

        assert output == expected_output


class TestChartJsSerializer:
    def test_output(self, collection: StaticCollection):
        serializer = serialize.ChartJsSerializer(
            collection,
            label_column="name",
            dataset_columns=["age"],
            dataset_labels={"age": "Age"},
            colors={"age": "test"},
        )

        output = serializer.serialize().strip()

        expected_output = json.dumps(
            {
                "labels": ["a", "b"],
                "datasets": [
                    {"data": [1, 2], "label": "Age", "backgroundColor": "test"},
                ],
            },
        )

        assert output == expected_output


class TestHtmlSerializer:
    @pytest.mark.usefixtures("with_request_context")
    def test_output(self, collection: StaticCollection, ckan_config: Any):
        serializer = serialize.HtmlSerializer(collection)
        output = serializer.render().strip()

        expected_output = (
            """<ul><li>{"age": 1, "name": "a"}</li>"""
            + """<li>{"age": 2, "name": "b"}</li></ul>"""
        )

        assert output == expected_output



from __future__ import annotations

import copy
import logging
from typing import Any, Callable, Iterable, Iterator

import ckan.plugins.toolkit as tk
from ckan.types import Context

from ckanext.collection import shared, types

from .base import Data

log = logging.getLogger(__name__)


class ApiData(Data[types.TData, types.TDataCollection], shared.UserTrait):
    """API data source.

    This base class is suitable for building API calls.

    Attributes:
      action: API action that returns the data
    """

    action: str = shared.configurable_attribute()
    payload: dict[str, Any] = shared.configurable_attribute(
        default_factory=lambda self: {},
    )
    ignore_auth: bool = shared.configurable_attribute(False)

    def make_context(self):
        return Context(user=self.user, ignore_auth=self.ignore_auth)

    def prepare_payload(self) -> dict[str, Any]:
        return copy.deepcopy(self.payload)

    def compute_data(self):
        action = tk.get_action(self.action)
        return action(self.make_context(), self.prepare_payload())


class ApiSearchData(ApiData[types.TData, types.TDataCollection]):
    """API data source optimized for package_search-like actions."""

    def prepare_payload(self) -> dict[str, Any]:
        payload = super().prepare_payload()
        return dict(
            payload,
            **self.get_filters(),
            **self.get_sort(),
        )

    def get_filters(self) -> dict[str, str]:
        return {}

    def get_sort(self) -> dict[str, str]:
        sort = self.attached.params.get("sort")
        if not sort:
            return {}

        column, desc = shared.parse_sort(sort)

        if column not in self.attached.columns.sortable:
            log.warning("Unexpected sort value: %s", sort)
            return {}

        direction = "desc" if desc else "asc"
        return {"sort": f"{column} {direction}"}

    def get_action(self) -> Callable[[Context, dict[str, Any]], Any]:
        return tk.get_action(self.action)

    def compute_data(self):
        action = self.get_action()
        return action(self.make_context(), dict(self.prepare_payload(), rows=0))

    def compute_total(self, data: dict[str, Any]) -> int:
        return data["count"]

    def range(self, start: int, end: int) -> Iterable[types.TData]:
        """@inherit"""
        action = self.get_action()
        return action(
            self.make_context(),
            dict(self.prepare_payload(), rows=end - start, start=start),
        )["results"]

    def __iter__(self) -> Iterator[types.TData]:
        action = self.get_action()
        context = self.make_context()
        start = 0
        while True:
            result = action(context, dict(self.prepare_payload(), start=start))

            yield from result["results"]
            start += len(result["results"])

            if start >= result["count"] or not result["results"]:
                break


class ApiListData(ApiSearchData[types.TData, types.TDataCollection]):
    """API data source optimized for organization_list-like actions."""

    def range(self, start: int, end: int) -> Iterable[types.TData]:
        """@inherit"""
        action = self.get_action()
        return action(
            self.make_context(),
            dict(self.prepare_payload(), limit=end - start, offset=start),
        )

    def compute_data(self):
        action = self.get_action()
        return action(self.make_context(), self.prepare_payload())

    def compute_total(self, data: dict[str, Any]) -> int:
        return len(data)

    def __iter__(self) -> Iterator[types.TData]:
        action = self.get_action()
        context = self.make_context()
        start = 0
        while True:
            result = action(context, dict(self.prepare_payload(), offset=start))
            if not result:
                break

            yield from result
            start += len(result)

            if start >= self.total:
                break



from __future__ import annotations

from typing import Any

import pytest

from ckan.tests.helpers import call_action

from ckanext.collection import config, shared, utils


class TestCollection:
    def test_basic_collection_creation(self):
        obj = utils.Collection("", {})
        assert isinstance(obj.columns, utils.Columns)
        assert isinstance(obj.pager, utils.ClassicPager)
        assert isinstance(obj.filters, utils.Filters)
        assert isinstance(obj.data, utils.Data)
        assert isinstance(obj.serializer, utils.Serializer)

    def test_custom_instance(self):
        data = utils.Data(None)

        collection = utils.Collection("", {}, data_instance=data)
        assert collection.data is data
        assert data.attached is collection

    def test_custom_factory(self):
        custom_factory = utils.Data[Any, Any].with_attributes()
        collection = utils.Collection[Any]("", {}, data_factory=custom_factory)

        assert isinstance(collection.data, custom_factory)

    def test_replace_service(self):
        collection = utils.Collection("", {})
        data = utils.Data(None)
        columns = utils.Columns(None)

        assert collection.data is not data
        assert collection.columns is not columns

        collection.replace_service(data)
        collection.replace_service(columns)

        assert collection.data is data
        assert collection.columns is columns


@pytest.mark.usefixtures("with_plugins")
class TestCollectionExplorer:
    def extract_names(self, collection: utils.Collection[Any]):
        return {
            f["value"]
            for f in collection.filters.filters[0]["options"]["options"]
            if f["value"]
        }

    @pytest.mark.ckan_config(config.CONFIG_ANNONYMOUS, "hello")
    def test_only_accessible_collections_are_visible(
        self,
        collection_registry: shared.Registry[Any],
    ):
        collection_registry.register("hello", utils.Collection)
        collection_registry.register("world", utils.Collection)

        explorer = utils.CollectionExplorer("", {})

        assert self.extract_names(explorer) == {"hello"}

        user = call_action("get_site_user")
        explorer = utils.CollectionExplorer(
            "",
            {},
            filters_settings={"user": user["name"]},
        )
        assert self.extract_names(explorer) == {"hello", "world"}



from __future__ import annotations

import abc
import csv
import io
import json
import operator
from functools import reduce
from typing import Any, Callable, Generic, Iterable, cast

import sqlalchemy as sa
from sqlalchemy.engine import Row
from sqlalchemy.exc import NoInspectionAvailable
from sqlalchemy.orm import InstanceState

import ckan.plugins.toolkit as tk

from ckanext.collection import shared, types

__all__ = [
    "Serializer",
    "StreamingSerializer",
    "RenderableSerializer",
    "CsvSerializer",
    "JsonlSerializer",
    "JsonSerializer",
    "ChartJsSerializer",
    "HtmlSerializer",
    "TableSerializer",
    "HtmxTableSerializer",
]


def basic_row_dictizer(row: Any) -> dict[str, Any]:
    if isinstance(row, dict):
        return cast("dict[str, Any]", row)

    if isinstance(row, Row):
        return dict(zip(row.keys(), row))

    try:
        reflection = sa.inspect(  # pyright: ignore[reportUnknownVariableType]
            row,
        )
    except NoInspectionAvailable:
        raise TypeError(type(row)) from None

    if isinstance(reflection, InstanceState):
        return {attr.key: attr.value for attr in reflection.attrs}  # pyright: ignore

    raise TypeError(type(row))


class Serializer(
    types.BaseSerializer,
    shared.Domain[types.TDataCollection],
    Generic[types.TSerialized, types.TDataCollection],
):
    """Base collection serializer.

    For any derived implementation, `serialize` must transfrom data of the
    collection into expected format. Example:

    >>> def stream(self):
    >>>     for record in self.attached.data:
    >>>         yield yaml.dump(record)
    >>>
    >>> def serialize(self):
    >>>     return "---\n".join(self.stream())

    """

    value_serializers: dict[str, types.ValueSerializer] = shared.configurable_attribute(
        default_factory=lambda self: {},
    )
    row_dictizer: Callable[[Any], dict[str, Any]] = shared.configurable_attribute(
        basic_row_dictizer,
    )

    def serialize(self) -> types.TSerialized:
        """Return serialized data."""
        return cast(types.TSerialized, None)

    def serialize_value(self, value: Any, name: str, record: Any):
        """Transform record's value into its serialized form."""
        for serializer_name, options in self.attached.columns.serializers.get(name, []):
            if serializer_name not in self.value_serializers:
                continue
            value = self.value_serializers[serializer_name](
                value,
                options,
                name,
                record,
                self,
            )

        return value

    def dictize_row(self, row: Any) -> dict[str, Any]:
        """Transform single data record into serializable dictionary."""
        result = self.row_dictizer(row)

        return {k: self.serialize_value(v, k, row) for k, v in result.items()}


class StreamingSerializer(Serializer[types.TSerialized, types.TDataCollection]):
    @abc.abstractmethod
    def stream(self) -> Iterable[types.TSerialized]:
        """Iterate over fragments of the content."""
        raise NotImplementedError

    def serialize(self):
        return reduce(operator.add, self.stream())


class RenderableSerializer(StreamingSerializer[str, types.TDataCollection]):
    def stream(self) -> Iterable[str]:
        """Iterate over fragments of the content."""
        yield ""

    def render(self) -> str:
        """Combine content fragments into a single dump."""
        return self.serialize()


class CsvSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into CSV document."""

    def get_writer(self, buff: io.StringIO):
        return csv.DictWriter(
            buff,
            fieldnames=self.get_fieldnames(),
            extrasaction="ignore",
        )

    def get_fieldnames(self) -> list[str]:
        return [
            name
            for name in self.attached.columns.names
            if name in self.attached.columns.visible
        ]

    def get_header_row(self, writer: csv.DictWriter[str]) -> dict[str, str]:
        return {
            col: self.attached.columns.labels.get(col, col) for col in writer.fieldnames
        }

    def prepare_row(self, row: Any, writer: csv.DictWriter[str]) -> dict[str, Any]:
        return self.dictize_row(row)

    def stream(self) -> Iterable[str]:
        buff = io.StringIO()
        writer = self.get_writer(buff)

        writer.writerow(self.get_header_row(writer))

        yield buff.getvalue()
        buff.seek(0)
        buff.truncate()

        for row in self.attached.data:
            writer.writerow(self.prepare_row(row, writer))
            yield buff.getvalue()
            buff.seek(0)
            buff.truncate()


class JsonlSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into JSONL lines."""

    def stream(self) -> Iterable[str]:
        buff = io.StringIO()
        for row in map(self.dictize_row, self.attached.data):
            json.dump(row, buff)
            yield buff.getvalue()
            yield "\n"
            buff.seek(0)
            buff.truncate()


class JsonSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into single JSON document."""

    def stream(self):
        yield json.dumps(
            [self.dictize_row(row) for row in self.attached.data],
        )


class ChartJsSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into data source for ChartJS module of
    ckanext-charts.

    """

    label_column: str = shared.configurable_attribute("")
    dataset_columns: list[str] = shared.configurable_attribute(
        default_factory=lambda self: [],
    )
    dataset_labels: dict[str, str] = shared.configurable_attribute(
        default_factory=lambda self: {},
    )
    colors: dict[str, str] = shared.configurable_attribute(
        default_factory=lambda self: {},
    )

    def stream(self):
        labels = []
        data: list[list[int]] = []

        for item in map(self.dictize_row, self.attached.data):
            labels.append(item.get(self.label_column, None))
            data.append(
                [item[name] for name in self.dataset_columns],
            )

        datasets: list[dict[str, Any]] = [
            {
                "data": row,
                "label": self.dataset_labels.get(column, column),
                "backgroundColor": self.colors.get(column),
            }
            for column, row in zip(self.dataset_columns, zip(*data))
        ]
        yield json.dumps(
            {
                "labels": labels,
                "datasets": datasets,
            },
        )


class HtmlSerializer(RenderableSerializer[types.TDataCollection]):
    """Serialize collection into HTML document."""

    ensure_dictized: bool = shared.configurable_attribute(False)

    main_template: str = shared.configurable_attribute(
        "collection/serialize/html/main.html",
    )
    record_template: str = shared.configurable_attribute(
        "collection/serialize/html/record.html",
    )

    def get_data(self) -> dict[str, Any]:
        return {
            "collection": self.attached,
        }

    def stream(self):
        yield tk.render(
            self.main_template,
            self.get_data(),
        )


class TableSerializer(HtmlSerializer[types.TDataCollection]):
    """Serialize collection into HTML table."""

    main_template: str = shared.configurable_attribute(
        "collection/serialize/table/main.html",
    )
    table_template: str = shared.configurable_attribute(
        "collection/serialize/table/table.html",
    )
    record_template: str = shared.configurable_attribute(
        "collection/serialize/table/record.html",
    )
    counter_template: str = shared.configurable_attribute(
        "collection/serialize/table/counter.html",
    )
    pager_template: str = shared.configurable_attribute(
        "collection/serialize/table/pager.html",
    )
    form_template: str = shared.configurable_attribute(
        "collection/serialize/table/form.html",
    )
    filter_template: str = shared.configurable_attribute(
        "collection/serialize/table/filter.html",
    )

    prefix: str = shared.configurable_attribute("collection-table")
    base_class: str = shared.configurable_attribute("collection")

    @property
    def form_id(self):
        return f"{self.prefix}-form--{self.attached.name}"

    @property
    def table_id(self):
        return f"{self.prefix}-id--{self.attached.name}"


class HtmxTableSerializer(TableSerializer[types.TDataCollection]):
    """Serialize collection into HTML table."""

    main_template: str = shared.configurable_attribute(
        "collection/serialize/htmx_table/main.html",
    )
    table_template: str = shared.configurable_attribute(
        "collection/serialize/htmx_table/table.html",
    )
    record_template: str = shared.configurable_attribute(
        "collection/serialize/htmx_table/record.html",
    )
    counter_template: str = shared.configurable_attribute(
        "collection/serialize/htmx_table/counter.html",
    )
    pager_template: str = shared.configurable_attribute(
        "collection/serialize/htmx_table/pager.html",
    )
    form_template: str = shared.configurable_attribute(
        "collection/serialize/htmx_table/form.html",
    )
    filter_template: str = shared.configurable_attribute(
        "collection/serialize/htmx_table/filter.html",
    )

    debug: bool = shared.configurable_attribute(False)
    push_url: bool = shared.configurable_attribute(False)
    base_class: str = shared.configurable_attribute("htmx-collection")

    render_url: str = shared.configurable_attribute(
        default_factory=lambda self: tk.h.url_for(
            "ckanext-collection.render",
            name=self.attached.name,
        ),
    )



from __future__ import annotations

import logging
from typing import Iterable

from ckanext.collection import shared, types

from .api import ApiData, ApiListData, ApiSearchData
from .base import Data
from .db import DbData, TableData
from .model import BaseSaData, ModelData, StatementSaData, UnionSaData

__all__ = [
    "Data",
    "TableData",
    "ApiData",
    "ApiListData",
    "ApiSearchData",
    "UnionSaData",
    "UnionModelData",
    "StatementSaData",
    "StatementModelData",
    "BaseSaData",
    "BaseModelData",
    "DbData",
    "ModelData",
]

BaseModelData = BaseSaData
UnionModelData = UnionSaData
StatementModelData = StatementSaData

log = logging.getLogger(__name__)


class StaticData(Data[types.TData, types.TDataCollection]):
    """Static data source.

    This class turns existing iterable into a data source.
    """

    data: Iterable[types.TData] = shared.configurable_attribute(
        default_factory=lambda self: [],
    )

    def compute_data(self) -> Iterable[types.TData]:
        return self.data



from __future__ import annotations

from typing import Any

import pytest
from faker import Faker

import ckan.plugins.toolkit as tk

from ckanext.collection import shared, utils


class AttachExample(shared.AttachTrait[Any]):
    pass


@pytest.mark.parametrize(
    "obj",
    [(None,), (1,), (object(),), ("hello",), (AttachExample(),)],
)
def test_attached_object(obj: Any):
    """Anything can be attached as a collection to AttachTrait implementation."""
    example = AttachExample()
    example._attach(obj)
    assert example.attached is obj


def test_service_attach_updates_collection():
    """Anything can be attached as a collection to AttachTrait implementation."""
    collection = utils.Collection("", {})
    example = AttachExample()
    example._attach(collection)
    assert collection.data is not example

    data = utils.Data(None)
    data._attach(collection)
    assert collection.data is data


class TestAttrSettings:
    def test_default_factories(self, faker: Faker):
        """Default factories are used when value is missing from settings."""
        name = "test"
        value = faker.pyint()
        another_value = faker.word()

        class Test(shared.AttrSettingsTrait):
            test = shared.configurable_attribute(default_factory=lambda self: value)

        obj = Test()

        obj._gather_settings({})
        assert getattr(obj, name) == value

        obj._gather_settings({name: another_value})
        assert getattr(obj, name) == another_value

    def test_without_factories(self, faker: Faker):
        """If defult factory is not specified, attribute is not created when
        it's missing from settings.

        """
        name = "test"
        value = faker.pyint()

        class Test(shared.AttrSettingsTrait):
            test = shared.configurable_attribute(None)

        obj = Test()

        obj._gather_settings({})
        assert getattr(obj, name) is None

        obj._gather_settings({name: value})
        assert getattr(obj, name) == value

    def test_declaration(self, faker: Faker):
        first = "first"
        second = "second"

        class Test(shared.AttrSettingsTrait):
            first = shared.configurable_attribute(default_factory=lambda self: "first")

        obj = Test()
        obj._gather_settings({})

        assert getattr(obj, first) == "first"

        class ChildTest(Test):
            first = shared.configurable_attribute(
                default_factory=lambda self: "new first",
            )

        obj = ChildTest()
        obj._gather_settings({})

        assert getattr(obj, first) == "new first"

        class AnotherChildTest(Test):
            second = shared.configurable_attribute(
                default_factory=lambda self: "second",
            )

        obj = AnotherChildTest()
        obj._gather_settings({})

        assert getattr(obj, first) == "first"
        assert getattr(obj, second) == "second"


@pytest.mark.usefixtures("clean_db")
def test_user_trait(user_factory: Any, faker: Faker, app_with_session: Any):
    fake_username = faker.name()

    class Test(shared.UserTrait):
        pass

    obj = Test()
    obj._gather_settings({})
    assert not obj.user

    with app_with_session.flask_app.test_request_context():
        user = user_factory.model()
        tk.login_user(user)

        obj._gather_settings({})
        assert obj.user == user.name

    obj._gather_settings({"user": fake_username})
    assert obj.user == fake_username


class TestDomain:
    def test_attachment(self):
        class Child(shared.Domain[Any]):
            ...

        attachment = object()
        obj = Child(attachment)
        assert obj.attached is attachment

    def test_settings(self):
        class Child(shared.Domain[Any]):
            prop = shared.configurable_attribute(
                default_factory=lambda self: self.attached,
            )

        obj = Child(object())
        assert obj.attached is obj.prop

        obj = Child(object(), prop=10)
        assert obj.prop == 10

    def test_with_attributes(self):
        """Domain.with_attributes creates anonymous class with overriden
        attributes.

        """

        class Child(shared.Domain[Any]):
            prop = shared.configurable_attribute(None)

        default = object()
        derived = Child.with_attributes(prop=shared.configurable_attribute(default))
        obj = derived(None)
        assert obj.prop is default

        obj = derived(None, prop=1)
        assert obj.prop == 1

        derived = Child.with_attributes(prop=default)
        obj = derived(None)
        assert obj.prop is default

        obj = derived(None, prop=1)
        assert obj.prop is default



from __future__ import annotations

from ckanext.collection.utils.data import ApiData, ApiListData, ApiSearchData

from .base import Collection


class ApiCollection(Collection):
    DataFactory = ApiData


class ApiSearchCollection(ApiCollection):
    DataFactory = ApiSearchData


class ApiListCollection(ApiCollection):
    DataFactory = ApiListData



from __future__ import annotations

from typing import Any
from unittest import mock

import pytest
import sqlalchemy as sa
from sqlalchemy.engine import Row

from ckan import model

from ckanext.collection.utils import Collection, data


@pytest.fixture()
def collection() -> Collection[Any]:
    return Collection("", {})


class TestLaziness:
    @mock.patch.object(data.Data, "compute_data", return_value=[])
    def test_data(self, stub: mock.Mock):
        obj = data.Data(None)
        assert not stub.called

        assert obj.total == 0
        assert stub.called

        stub.reset_mock()

        obj = data.Data(None)
        assert not stub.called

        assert list(obj) == []
        assert stub.called

        stub.reset_mock()

        obj = data.Data(None)
        assert list(obj) == []
        assert len(obj) == 0
        assert list(obj) == []
        assert len(obj) == 0

        stub.assert_called_once()

    @mock.patch.object(data.Data, "compute_total", return_value=0)
    def test_total(self, stub: mock.Mock):
        obj = data.Data(None)
        assert not stub.called

        assert obj.total == 0
        assert stub.called

        stub.reset_mock()

        obj = data.Data(None)
        assert not stub.called

        assert obj._data == []
        assert not stub.called

        stub.reset_mock()

        obj = data.Data(None)
        assert list(obj) == []
        assert len(obj) == 0
        assert list(obj) == []
        assert len(obj) == 0

        stub.assert_called_once()


class TestStaticData:
    def test_settings(self, collection: Collection[Any]):
        dump = [{"a": 1}, {"a": 2}]
        obj = data.StaticData(collection, data=dump)

        assert list(obj) == dump
        assert len(obj) == len(dump)

    def test_pagination(self):
        dump = [{"a": 1}, {"a": 2}]

        collection = Collection[Any](
            "",
            {"page": 2, "rows_per_page": 1},
            data_settings={"data": dump},
            data_factory=data.StaticData,
        )

        assert list(collection) == [dump[1]]
        assert len(collection.data) == 2


@pytest.mark.usefixtures("clean_db")
class TestModelData:
    def test_empty_result(self, collection: Collection[Any]):
        obj = data.ModelData(collection, model=model.Package)
        assert obj.total == 0

    def test_non_empty_result(self, package_factory: Any, collection: Collection[Any]):
        package_factory.create_batch(3)
        obj = data.ModelData(collection, model=model.Package)
        assert isinstance(list(obj)[0], Row)
        assert obj.total == 3
        assert len(list(obj)) == 3

    def test_scalar(self, package: Any, collection: Collection[Any]):
        obj = data.ModelData(collection, model=model.Package)
        assert isinstance(list(obj)[0], Row)

        obj = data.ModelData(collection, model=model.Package, is_scalar=True)
        assert isinstance(list(obj)[0], model.Package)

    def test_columns(self, package: Any, collection: Collection[Any]):
        obj = data.ModelData(collection, model=model.Package)
        row = next(iter(obj))
        assert hasattr(row, "id")
        assert hasattr(row, "name")

        obj = data.ModelData(
            collection,
            model=model.Package,
            static_columns=[model.Package.id],
        )
        row = next(iter(obj))
        assert hasattr(row, "id")
        assert not hasattr(row, "name")

    def test_joins(self, package_factory: Any, collection: Collection[Any]):
        parent = package_factory()
        child = package_factory(notes=parent["id"])

        child_alias = sa.alias(model.Package)
        obj = data.ModelData(
            collection,
            model=model.Package,
            static_sources={"child": child_alias},
            static_columns=[
                model.Package.name.label("parent"),
                child_alias.c.name.label("child"),
            ],
            static_joins=[("child", child_alias.c.notes == model.Package.id, False)],
        )
        row = next(iter(obj))
        assert row.parent == parent["name"]
        assert row.child == child["name"]

    def test_filters(self, package_factory: Any, collection: Collection[Any]):
        pkg = package_factory(notes="x")
        package_factory(notes="y")

        obj = data.ModelData(
            collection,
            model=model.Package,
            static_filters=[model.Package.notes == "x"],
        )
        assert obj.total == 1
        assert next(iter(obj)).id == pkg["id"]

    def test_sorting(self, package_factory: Any):
        ids = sorted([pkg["id"] for pkg in package_factory.create_batch(3)])

        collection = Collection("", {}, columns_settings={"sortable": {"id"}})
        obj = data.ModelData(
            collection,
            model=model.Package,
            is_scalar=True,
            static_columns=[model.Package.id],
        )

        collection.params["sort"] = "id"
        obj.refresh_data()
        assert list(obj) == ids

        collection.params["sort"] = "-id"
        obj.refresh_data()
        assert list(obj) == list(reversed(ids))

        collection.params["sort"] = "id asc"
        obj.refresh_data()
        assert list(obj) == ids

        collection.params["sort"] = "id desc"
        obj.refresh_data()
        assert list(obj) == list(reversed(ids))

    def test_limits(self, package_factory: Any):
        ids = sorted([pkg["id"] for pkg in package_factory.create_batch(3)])

        collection = Collection(
            "",
            {"sort": "id"},
            pager_settings={"page": 2, "rows_per_page": 1, "sort": "id"},
            data_factory=data.ModelData,
            columns_settings={"sortable": "id"},
            data_settings={
                "model": model.Package,
                "is_scalar": True,
                "static_columns": [model.Package.id],
            },
        )

        assert list(collection) == [ids[1]]


@pytest.mark.usefixtures("clean_db", "clean_index")
class TestApiSearchData:
    def test_base(self, package_factory: Any):
        ids = sorted([pkg["id"] for pkg in package_factory.create_batch(3)])

        collection = Collection("", {})
        obj = data.ApiSearchData(collection, action="package_search")

        assert obj.total == 3
        assert sorted([p["id"] for p in obj]) == ids

        obj = data.ApiSearchData(
            collection,
            action="package_search",
            payload={"fq": "id:" + ids[1]},
        )

        assert obj.total == 1
        assert next(iter(obj))["id"] == ids[1]


@pytest.mark.usefixtures("clean_db")
class TestApiListData:
    def test_base(self, organization_factory: Any):
        ids = sorted([o["name"] for o in organization_factory.create_batch(3)])

        collection = Collection("", {})
        obj = data.ApiListData(collection, action="organization_list")

        assert obj.total == 3
        assert sorted(obj) == ids

    def test_payload(self, organization_factory: Any):
        organization_factory(type="custom")

        collection = Collection("", {})

        obj = data.ApiListData(collection, action="organization_list")
        assert obj.total == 0

        obj = data.ApiListData(
            collection,
            action="organization_list",
            payload={"type": "custom"},
        )
        assert obj.total == 1



from __future__ import annotations

import abc
import logging
from functools import cached_property
from typing import Any, Generic, Iterable, Iterator, TypeVar, cast

import sqlalchemy as sa
from sqlalchemy.orm import Mapper
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.elements import Label
from sqlalchemy.sql.selectable import GenerativeSelect, Select

from ckan import model
from ckan.types import AlchemySession

from ckanext.collection import shared, types

from .base import Data

log = logging.getLogger(__name__)
TStatement = TypeVar("TStatement", bound=GenerativeSelect)


class BaseSaData(
    Data[types.TData, types.TDataCollection],
    Generic[TStatement, types.TData, types.TDataCollection],
):
    """Data source for custom SQL statement."""

    _data: cached_property[TStatement]
    use_naive_filters: bool = shared.configurable_attribute(False)
    use_naive_search: bool = shared.configurable_attribute(False)
    session: AlchemySession = shared.configurable_attribute(
        default_factory=lambda self: model.Session,
    )

    def compute_data(self):
        stmt = self.get_base_statement()
        stmt = self.alter_statement(stmt)
        stmt = self.statement_with_filters(stmt)
        return self.statement_with_sorting(stmt)

    @abc.abstractmethod
    def get_base_statement(self) -> Any:
        """Return statement with minimal amount of columns and filters."""
        ...

    def compute_total(self, data: TStatement) -> int:
        return self.count_statement(data)

    def __iter__(self) -> Iterator[types.TData]:
        yield from self.execute_statement(self._data)

    def _execute(self, stmt: GenerativeSelect):
        return self.session.execute(stmt)

    def range(self, start: int, end: int) -> Iterable[types.TData]:
        stmt = self._data.limit(end - start).offset(start)
        return self.execute_statement(stmt)

    def execute_statement(self, stmt: TStatement) -> Iterable[types.TData]:
        result: Any = self._execute(stmt)
        return result

    def alter_statement(self, stmt: TStatement):
        """Add columns, joins and unconditional filters to statement."""
        return stmt

    def count_statement(self, stmt: TStatement) -> int:
        """Count number of items in query"""
        count_stmt: Select = sa.select(sa.func.count()).select_from(stmt)
        return cast(int, self._execute(count_stmt).scalar())

    def _into_clause(self, column: ColumnElement[Any], value: Any):
        if isinstance(value, list):
            return column.in_(value)

        return column == value

    def statement_with_filters(self, stmt: TStatement) -> TStatement:
        """Add normal filter to statement."""

        if not isinstance(stmt, Select):
            return stmt

        params = self.attached.params
        if self.use_naive_filters:
            stmt = stmt.where(
                sa.and_(
                    sa.true(),
                    *[
                        self._into_clause(stmt.selected_columns[name], params[name])
                        for name in self.attached.columns.filterable
                        if name in params
                        and params[name] != ""
                        and name in stmt.selected_columns
                    ],
                ),
            )

        if self.use_naive_search:
            if (q := params.get("q")) and "q" not in self.attached.columns.searchable:
                stmt = stmt.where(
                    sa.or_(
                        sa.false(),
                        *[
                            stmt.selected_columns[name].ilike(f"%{q}%")
                            for name in self.attached.columns.searchable
                            if name in stmt.selected_columns
                        ],
                    ),
                )
            for name in self.attached.columns.searchable:
                if name not in params or name not in stmt.selected_columns:
                    continue
                value = params[name]
                stmt = stmt.where(stmt.selected_columns[name].ilike(f"%{value}%"))

        return stmt

    def statement_with_sorting(self, stmt: TStatement):
        """Specify sorting order for statement."""
        sort = self.attached.params.get("sort")
        if not sort:
            return stmt

        column, desc = shared.parse_sort(sort)

        if column not in self.attached.columns.sortable:
            log.warning("Unexpected sort value: %s", column)
            return stmt

        if column not in stmt.selected_columns:
            sort = column
            if desc:
                sort = f"{sort} DESC"
            return stmt.order_by(sa.text(sort))

        col_object = stmt.selected_columns[column]

        return stmt.order_by(col_object.desc() if desc else col_object.asc())


class StatementSaData(BaseSaData[Select, types.TData, types.TDataCollection]):
    """Data source for custom SQL statement."""

    statement: Any = shared.configurable_attribute(None)

    def get_base_statement(self):
        """Return statement with minimal amount of columns and filters."""
        return self.statement


class UnionSaData(BaseSaData[Select, types.TData, types.TDataCollection]):
    """Data source for custom SQL statement."""

    statements: Iterable[GenerativeSelect] = shared.configurable_attribute(
        default_factory=lambda self: [],
    )

    def get_base_statement(self):
        """Return statement with minimal amount of columns and filters."""
        return sa.select(sa.union_all(*self.statements).subquery())


class ModelData(BaseSaData[Select, types.TData, types.TDataCollection]):
    """DB data source.

    This base class is suitable for building SQL query.

    Attributes:
      model: main model used by data source
      is_scalar: return model instance instead of columns set.
    """

    model: Any = shared.configurable_attribute(None)
    is_scalar: bool = shared.configurable_attribute(False)

    static_columns: list[sa.Column[Any] | Label[Any]] = shared.configurable_attribute(
        default_factory=lambda self: [],
    )
    static_filters: list[Any] = shared.configurable_attribute(
        default_factory=lambda self: [],
    )
    static_sources: dict[str, Any] = shared.configurable_attribute(
        default_factory=lambda self: {},
    )
    static_joins: list[tuple[str, Any, bool]] = shared.configurable_attribute(
        default_factory=lambda self: [],
    )

    def execute_statement(self, stmt: Select) -> Iterable[types.TData]:
        result: Any
        if self.is_scalar:
            result = self.session.scalars(stmt)

        else:
            result = self.session.execute(stmt)

        return result

    def select_columns(self) -> Iterable[Any]:
        """Return list of columns for select statement."""
        if self.static_columns:
            return self.static_columns

        if not self.model:
            return []

        mapper = cast(Mapper, sa.inspect(self.model))
        return [self.model] if self.is_scalar else [mapper.columns]

    def get_extra_sources(self) -> dict[str, Any]:
        """Return mapping of additional models/subqueries used to build the
        statement.

        Note: Don't call this method direclty. Instead, use `extra_sources`
        property, that caches return value of the current method. Extra sources
        must not change during query building process.

        Extra sources are used when the query contains subqueries or joins with
        aliased models. For example, in order to select column from joined
        subquery, instead of creating identical subqueries inside
        `select_columns` and `get_joins`, write the following code:

        >>> def get_extra_sources(self):
        >>>     return {"subq": make_subquery()}
        >>>
        >>> def select_columns(self):
        >>>     return [self.extra_sources["subq"].c.column_name]
        >>>
        >>> def get_joins(self):
        >>>     subq = self.extra_sources["subq"]
        >>>     return [("subq", subq.c.id == self.model.id, False)]

        """
        return self.static_sources

    @cached_property
    def extra_sources(self) -> dict[str, Any]:
        """Cache for extra sources.

        If you are going to override this property, don't forget to add
        `cached_property` decorator. The better option would be to override
        `get_extra_sources`. In this way you will not be affected if cache
        implementation changes in future.
        """
        return self.get_extra_sources()

    def get_joins(self) -> Iterable[tuple[str, Any, bool]]:
        """Return join details for extra_sources.

        Every item is a tuple of of (extra_source_name, join_condition,
        outerjoin_flag)
        """
        return self.static_joins

    def apply_joins(self, stmt: Select) -> Select:
        """Return list of columns for select statement."""

        sources = self.extra_sources

        for name, condition, isouter in self.get_joins():
            stmt = stmt.join(sources[name], condition, isouter)

        return stmt

    def get_base_statement(self):
        """Return statement with minimal amount of columns and filters."""
        columns = self.select_columns()
        stmt = sa.select(*columns)
        if self.model:
            stmt = stmt.select_from(self.model)

        return self.apply_joins(stmt)

    def statement_with_filters(self, stmt: Select):
        """Add normal filter to statement."""
        for cond in self.static_filters:
            stmt = stmt.where(cond)

        return super().statement_with_filters(stmt)



from __future__ import annotations

from functools import cached_property
from typing import Any, Generic, Iterable, Iterator

from ckanext.collection import shared, types


class Data(
    types.BaseData,
    shared.Domain[types.TDataCollection],
    Generic[types.TData, types.TDataCollection],
):
    """Data source for collection.

    This class produces data for collection.

    """

    def __init__(self, obj: types.TDataCollection, /, **kwargs: Any):
        super().__init__(obj, **kwargs)

    def __iter__(self) -> Iterator[types.TData]:
        yield from self._data

    def __len__(self):
        return self.total

    def refresh_data(self):
        """Update data and counters.

        Use this method when parameters that affect Data are changed. This is a
        cheaper alternative of recreating the whole collection with new
        parameters.

        """
        self._data = self.compute_data()
        self._total = self.compute_total(self._data)

    def compute_data(self) -> Any:
        """Produce data."""
        return []

    def compute_total(self, data: Any) -> int:
        """Compute number of data records."""
        return len(data)

    def range(self, start: Any, end: Any) -> Iterable[types.TData]:
        """Slice data."""
        return self._data[start:end]

    @property
    def total(self):
        return self._total

    @cached_property
    def _data(self):
        return self.compute_data()

    @cached_property
    def _total(self) -> int:
        return self.compute_total(self._data)



from __future__ import annotations

import pytest

from ckanext.collection.utils import Collection, columns


@pytest.fixture()
def collection():
    return Collection("", {})


class TestColumns:
    def test_defaults(self, collection: Collection):
        obj = columns.Columns(collection)
        assert obj.names == []
        assert obj.hidden == set()
        assert obj.visible == set()
        assert obj.sortable == set()
        assert obj.filterable == set()
        assert obj.labels == {}

    def test_names(self, collection: Collection):
        names = ["a", "b", "c"]
        obj = columns.Columns(collection, names=names)
        assert obj.names == names
        assert obj.hidden == set()
        assert obj.visible == set(names)
        assert obj.sortable == set(names)
        assert obj.filterable == set(names)
        assert obj.labels == {n: n for n in names}

    def test_hidden(self, collection: Collection):
        names = ["a", "b", "c"]
        obj = columns.Columns(collection, names=names, hidden={"b"})
        assert obj.names == names
        assert obj.hidden == {"b"}
        assert obj.visible == {"a", "c"}



from __future__ import annotations

from typing import Any, Iterator, overload

from typing_extensions import Self

from ckanext.collection import shared, types
from ckanext.collection.utils.columns import Columns
from ckanext.collection.utils.data import Data
from ckanext.collection.utils.filters import Filters
from ckanext.collection.utils.pager import ClassicPager, Pager
from ckanext.collection.utils.serialize import Serializer


class Collection(types.BaseCollection):
    """Base data collection.

    Contains the following information:
    * data itself
    * available filters
    * details about table columns
    * pagination details
    * serializer instance

    In most scenarious, you'll create a subclass of the collection with
    different factories. If you have `ApiData` subclass of `Data`, which
    produces data using CKAN API, you can create the following collection
    subclass:

    >>> class ApiCollection(Collection):
    >>>     DataFactory = ApiData

    Collections can be created on demand, using constructor parameters. For
    example, ApiCollection mentioned above can be instantiated even without
    dedicated class:

    >>> api_collection = Collection(
    >>>     data_factory=ApiData,
    >>> )

    Subclasses are more flexible, than dynamic collections. They allow
    overriding and enhancing different aspects of collection. But creating
    collections on demand may be more convenient if you need a number of
    similar collections and prefer not to bloat the codebase with almost
    identical classes.

    Apart from mandatory `name` and `params` arguments, all arguments to the
    constructor must be named and follow one of three patterns:

    * SMTH_factory - class used to instantiate an attribute. Example:

      >>> Colection(data_factory=ModelData)
      >>> Colection(serializer_factory=JsonSerializer)

    * SMTH_instance - object assigned to the attribute. If present, factory is
      ignored. Example:

      >>> col = Colection()
      >>> another_col = Colection(pager_instance=col.pager)

    * SMTH_settings - collection of named arguments passed into constructor of
      the corresponding attribute. Example

      >>> Collection(pager_settings={"rows_per_page": 100})
      >>> Collection(columns_settings={"names": ["title", "notes"]})

    Attributes:
      name: unique name of the collection
      params: data used for search/pagination/sorting/etc.

    """

    # keep these classes here to simplify overrides
    ColumnsFactory: type[Columns[Self]] = Columns
    DataFactory: type[Data[Any, Self]] = Data
    FiltersFactory: type[Filters[Self]] = Filters
    SerializerFactory: type[Serializer[Any, Self]] = Serializer
    PagerFactory: type[Pager[Self]] = ClassicPager

    _service_names: tuple[str, ...] = (
        "columns",
        "pager",
        "filters",
        "data",
        "serializer",
    )

    def __init__(self, name: str, params: dict[str, Any], /, **kwargs: Any):
        """Use name to pick only relevant parameters.

        When multiple collections rendered on the same page, the use format
        `collection:param` to avoid conflicts with parameters from different
        collections.

        """
        self.name = name

        if self.name:
            params = {
                k[len(self.name) + 1 :]: v
                for k, v in params.items()
                if k.startswith(f"{self.name}:")
            }
        self.params = params

        for service in self._service_names:
            self.replace_service(self._instantiate(service, kwargs))

    def _instantiate(self, name: str, kwargs: dict[str, Any]) -> Any:
        if factory := kwargs.get(f"{name}_factory"):
            fn = "".join(p.capitalize() for p in name.split("_"))
            setattr(self, fn + "Factory", factory)

        value: shared.Domain[Any] | None = kwargs.get(f"{name}_instance")
        if value is None:
            maker = getattr(self, f"make_{name}")
            value = maker(**kwargs.get(f"{name}_settings", {}))

        else:
            value._attach(self)  # pyright: ignore [reportPrivateUsage]

        return value

    @overload
    def replace_service(self, service: types.BaseColumns) -> types.BaseColumns | None:
        ...

    @overload
    def replace_service(self, service: types.BaseData) -> types.BaseData | None:
        ...

    @overload
    def replace_service(self, service: types.BasePager) -> types.BasePager | None:
        ...

    @overload
    def replace_service(self, service: types.BaseFilters) -> types.BaseFilters | None:
        ...

    @overload
    def replace_service(
        self,
        service: types.BaseSerializer,
    ) -> types.BaseSerializer | None:
        ...

    @overload
    def replace_service(
        self,
        service: types.BaseDbConnection,
    ) -> types.BaseDbConnection | None:
        ...

    def replace_service(self, service: types.Service) -> types.Service | None:
        """Attach service to collection"""
        old_service = getattr(self, service.service_name, None)
        setattr(self, service.service_name, service)
        return old_service

    def __iter__(self) -> Iterator[Any]:
        yield from self.data.range(self.pager.start, self.pager.end)

    def make_serializer(self, **kwargs: Any) -> Serializer[Any, Self]:
        """Return serializer."""
        return self.SerializerFactory(self, **kwargs)

    def make_pager(self, **kwargs: Any) -> Pager[Self]:
        """Return pager."""
        return self.PagerFactory(self, **kwargs)

    def make_columns(self, **kwargs: Any) -> Columns[Self]:
        """Return column details."""
        return self.ColumnsFactory(self, **kwargs)

    def make_filters(self, **kwargs: Any) -> Filters[Self]:
        """Return search filters."""
        return self.FiltersFactory(self, **kwargs)

    def make_data(self, **kwargs: Any) -> Data[Any, Self]:
        """Return search filters."""
        return self.DataFactory(self, **kwargs)



from __future__ import annotations

from flask import Blueprint
from werkzeug.utils import secure_filename

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.common import streaming_response
from ckan.logic import parse_params

from ckanext.collection import config, shared
from ckanext.collection.utils.serialize import StreamingSerializer

bp = Blueprint("ckanext-collection", __name__)


try:
    from ckanext.ap_main.views.generics import ApConfigurationPageView

    class ApConfiguration(ApConfigurationPageView):
        pass

    bp.add_url_rule(
        "/admin-panel/config/collection",
        view_func=ApConfiguration.as_view("ap_config", "ckanext-collection-ap-config"),
    )

except ImportError:
    pass


@bp.route("/api/util/collection/<name>/render")
def render(name: str) -> str | bytes:
    try:
        tk.check_access("collection_view_render", {}, {"name": name})
    except tk.NotAuthorized:
        return tk.abort(403)

    params = parse_params(tk.request.args)

    collection = shared.get_collection(name, params)

    if not collection:
        return tk.abort(404)

    return collection.serializer.serialize()


@bp.route("/api/util/collection/<name>/export")
@bp.route("/api/util/collection/<name>/export/<format>")
def export(name: str, format: str | None = None) -> types.Response:
    try:
        tk.check_access("collection_view_export", {}, {"name": name})
    except tk.NotAuthorized:
        return tk.abort(403)

    params = parse_params(tk.request.args)

    collection = shared.get_collection(name, params)
    if not collection:
        return tk.abort(404)

    if format:
        serializer_factory = config.serializer(format)
        if not serializer_factory:
            return tk.abort(404)
        serializer = serializer_factory(collection)
        filename = f"collection.{format}"

    else:
        serializer = collection.serializer
        filename = "collection"

    if not isinstance(serializer, StreamingSerializer):
        return tk.abort(409, "This collection does not support streaming export")

    filename = secure_filename(tk.request.args.get("filename", filename))

    resp = streaming_response(serializer.stream(), with_context=True)
    resp.headers["Content-Disposition"] = f'attachment; filename="{filename}"'
    return resp



from __future__ import annotations

import operator
from dataclasses import is_dataclass
from typing import Any

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan.common import CKANConfig

from . import shared
from .interfaces import CollectionFactory, ICollection

try:
    from ckanext.ap_main.interfaces import IAdminPanel
    from ckanext.ap_main.types import ConfigurationItem, SectionConfig

    class ApImplementation(p.SingletonPlugin):  # pyright: ignore[reportRedeclaration]
        p.implements(IAdminPanel, inherit=True)

        def register_config_sections(
            self,
            config_list: list[SectionConfig],
        ) -> list[SectionConfig]:
            """Extension will receive the list of section config objects."""
            config_page = ConfigurationItem(
                name="Collections",
                info="ckanext-collection configuration",
                blueprint="ckanext-collection.ap_config",
            )
            getter: Any = (
                operator.attrgetter
                if is_dataclass(SectionConfig)
                else operator.itemgetter
            )

            for section in config_list:
                if getter("name")(section) == "Basic site settings":
                    getter("configs")(section).append(config_page)
                    break
            else:
                config_list.append(
                    SectionConfig(  # type: ignore
                        name="Basic site settings",
                        configs=[config_page],
                    ),
                )

            return config_list

except ImportError:

    class ApImplementation(p.SingletonPlugin):
        pass


register_collection_signal = tk.signals.ckanext.signal(
    "collection:register_collections",
)


@tk.blanket.blueprints
@tk.blanket.auth_functions
@tk.blanket.config_declarations
@tk.blanket.helpers
@tk.blanket.cli
class CollectionPlugin(ApImplementation, p.SingletonPlugin):
    p.implements(p.IConfigurer)
    p.implements(p.IConfigurable)
    # p.implements(p.ISignal)
    p.implements(ICollection, inherit=True)

    # IConfigurer

    def update_config(self, config_: CKANConfig):
        tk.add_template_directory(config_, "templates")
        tk.add_public_directory(config_, "public")
        tk.add_resource("assets", "collection")

    # IConfigurable
    def configure(self, config_: CKANConfig):
        shared.collection_registry.reset()

        for plugin in p.PluginImplementations(ICollection):
            for name, factory in plugin.get_collection_factories().items():
                shared.collection_registry.register(name, factory)

        for _, factories in register_collection_signal.send():
            for name, factory in factories.items():
                shared.collection_registry.register(name, factory)

    def get_collection_factories(self) -> dict[str, CollectionFactory]:
        if tk.config["debug"]:
            from .utils import (
                CkanDbConnection,
                CollectionExplorer,
                DatastoreDbConnection,
                DbExplorer,
            )

            return {
                "collection-explorer": CollectionExplorer,
                "core-db-explorer": lambda n, p, **kwargs: DbExplorer(
                    n,
                    p,
                    **dict(kwargs, db_connection_factory=CkanDbConnection),
                ),
                "datastore-db-explorer": lambda n, p, **kwargs: DbExplorer(
                    n,
                    p,
                    **dict(kwargs, db_connection_factory=DatastoreDbConnection),
                ),
            }

        return {}



from __future__ import annotations


from ckan.plugins import Interface

from ckanext.collection.types import CollectionFactory


class ICollection(Interface):
    """Extend functionality of ckanext-collections"""

    def get_collection_factories(self) -> dict[str, CollectionFactory]:
        """Register named collection factories.

        Example:
            def get_collection_factories(self) -> dict[str, CollectionFactory]:
                return {
                    "packages": PackageCollection,
                }
        """
        return {}



import pytest

from ckanext.collection import shared


@pytest.fixture()
def collection_registry():
    """Collection registry cleaned after each test."""

    yield shared.collection_registry
    shared.collection_registry.reset()



from __future__ import annotations

from typing import Any

import ckan.plugins.toolkit as tk

from ckanext.collection import shared, types


class Pager(types.BasePager, shared.Domain[types.TDataCollection]):
    """Pagination logic for collections.

    This class must be abstract enough to fit into majority of pager
    implementations.

    """


class ClassicPager(Pager[types.TDataCollection]):
    """Limit/offset based pagination.

    Attributes:
      page: current active page
      rows_per_page: max number of items per page
    """

    page: int = shared.configurable_attribute(1)
    rows_per_page: int = shared.configurable_attribute(10)

    def __init__(self, obj: types.TDataCollection, /, **kwargs: Any):
        """Use `page` and `rows_per_page` parameters."""
        super().__init__(obj, **kwargs)
        self.page = tk.h.get_page_number(self.attached.params, "page", self.page)
        self.rows_per_page = tk.h.get_page_number(
            self.attached.params,
            "rows_per_page",
            self.rows_per_page,
        )

    @property
    def start(self) -> int:
        return self.size * self.page - self.size

    @property
    def end(self) -> int:
        return self.start + self.size

    @property
    def size(self) -> Any:
        return self.rows_per_page
